import numpy as np
from typing import Tuple, List
from textalloc.candidates import generate_candidates
from textalloc.overlap_functions import (
    non_overlapping_with_points,
    non_overlapping_with_lines,
    non_overlapping_with_boxes,
    inside_plot,
)
from tqdm import tqdm


def get_non_overlapping_boxes(
    original_boxes: np.ndarray,
    xlims: Tuple[float, float],
    ylims: Tuple[float, float],
    distance_margin_fraction: float,
    verbose: bool,
    nbr_candidates: int,
    scatter_xy: np.ndarray = None,
    lines_xyxy: np.ndarray = None,
) -> Tuple[List[Tuple[float, float, float, float, str, int]], List[int]]:
    """Finds boxes that do not have an overlap with any other objects.

    Args:
        original_boxes (np.ndarray): original boxes containing texts.
        xlims (Tuple[float, float]): x-limits of plot gotten from ax.get_ylim()
        ylims (Tuple[float, float]): y-limits of plot gotten from ax.get_ylim()
        distance_margin_fraction (float): parameter for margins between objects. Increase for larger margins to points and lines.
        verbose (bool): prints progress using tqdm.
        nbr_candidates (int): Sets the number of candidates used.
        scatter_xy (np.ndarray, optional): 2d array of scattered points in plot. Defaults to None.
        lines_xyxy (np.ndarray, optional): 2d array of line segments in plot.. Defaults to None.

    Returns:
        Tuple[List[Tuple[float, float, float, float, str, int]], List[int]]: data of non-overlapping boxes and indices of overlapping boxes.
    """
    xmin_bound, xmax_bound = xlims
    ymin_bound, ymax_bound = ylims

    xfrac = (xmax_bound - xmin_bound) * distance_margin_fraction
    yfrac = (ymax_bound - ymin_bound) * distance_margin_fraction

    box_arr = np.zeros((0, 4))
    non_overlapping_boxes = []

    # Iterate original boxes and find ones that do not overlap by creating multiple candidates
    non_overlapping_boxes = []
    overlapping_boxes_inds = []
    for i, box in tqdm(enumerate(original_boxes), disable=not verbose):
        x_original, y_original, w, h, s = box
        candidates = generate_candidates(
            w,
            h,
            x_original,
            y_original,
            xfrac / 2,
            yfrac / 2,
            nbr_candidates=nbr_candidates,
        )

        # Check for overlapping
        if scatter_xy is None:
            non_op = np.zeros((candidates.shape[0],)) == 0
        else:
            non_op = non_overlapping_with_points(
                scatter_xy, candidates, xfrac / 2, yfrac / 2
            )
        if lines_xyxy is None:
            non_ol = np.zeros((candidates.shape[0],)) == 0
        else:
            non_ol = non_overlapping_with_lines(
                lines_xyxy, candidates, xfrac / 2, yfrac / 2
            )
        if box_arr.shape[0] == 0:
            non_orec = np.zeros((candidates.shape[0],)) == 0
        else:
            non_orec = non_overlapping_with_boxes(
                box_arr, candidates, xfrac / 2, yfrac / 2
            )
        inside = inside_plot(xmin_bound, ymin_bound, xmax_bound, ymax_bound, candidates)

        # Validate
        ok_candidates = np.where(
            np.bitwise_and(
                non_ol, np.bitwise_and(non_op, np.bitwise_and(non_orec, inside))
            )
        )[0]
        if len(ok_candidates) > 0:
            best_candidate = candidates[ok_candidates[0], :]
            box_arr = np.vstack(
                [
                    box_arr,
                    np.array(
                        [
                            best_candidate[0],
                            best_candidate[1],
                            best_candidate[0] + w,
                            best_candidate[1] + h,
                        ]
                    ),
                ]
            )
            non_overlapping_boxes.append(
                (best_candidate[0], best_candidate[1], w, h, s, i)
            )
        else:
            overlapping_boxes_inds.append(i)
    return non_overlapping_boxes, overlapping_boxes_inds



import numpy as np


def non_overlapping_with_points(
    scatter_xy: np.ndarray, candidates: np.ndarray, xfrac: float, yfrac: float
) -> np.ndarray:
    """Finds candidates not overlapping with points.

    Args:
        scatter_xy (np.ndarray): Array of shape (N,2) containing coordinates for all scatter-points
        candidates (np.ndarray): Array of shape (K,4) with K candidate boxes
        xfrac (float): fraction of the x-dimension to use as margins for text boxes
        yfrac (float): fraction of the y-dimension to use as margins for text boxes

    Returns:
        np.ndarray: Boolean array of shape (K,) with True for non-overlapping candidates with points
    """
    return np.invert(
        np.bitwise_or.reduce(
            np.bitwise_and(
                candidates[:, 0][:, None] - xfrac < scatter_xy[:, 0],
                np.bitwise_and(
                    candidates[:, 2][:, None] + xfrac > scatter_xy[:, 0],
                    np.bitwise_and(
                        candidates[:, 1][:, None] - yfrac < scatter_xy[:, 1],
                        candidates[:, 3][:, None] + yfrac > scatter_xy[:, 1],
                    ),
                ),
            ),
            axis=1,
        )
    )


def non_overlapping_with_lines(
    lines_xyxy: np.ndarray, candidates: np.ndarray, xfrac: float, yfrac: float
) -> np.ndarray:
    """Finds candidates not overlapping with lines

    Args:
        lines_xyxy (np.ndarray): line segments
        candidates (np.ndarray): candidate boxes
        xfrac (float): fraction of the x-dimension to use as margins for text boxes
        yfrac (float): fraction of the y-dimension to use as margins for text boxes

    Returns:
        np.ndarray: Boolean array of shape (K,) with True for non-overlapping candidates with lines.
    """
    non_intersecting = np.invert(
        np.any(
            np.bitwise_or(
                line_intersect(
                    np.hstack(
                        [
                            candidates[:, 0:1] - xfrac,
                            candidates[:, 1:2] - yfrac,
                            candidates[:, 0:1] - xfrac,
                            candidates[:, 3:] + yfrac,
                        ]
                    ),
                    lines_xyxy,
                ),
                np.bitwise_or(
                    line_intersect(
                        np.hstack(
                            [
                                candidates[:, 0:1] - xfrac,
                                candidates[:, 3:] + yfrac,
                                candidates[:, 2:3] + xfrac,
                                candidates[:, 3:] + yfrac,
                            ]
                        ),
                        lines_xyxy,
                    ),
                    np.bitwise_or(
                        line_intersect(
                            np.hstack(
                                [
                                    candidates[:, 2:3] + xfrac,
                                    candidates[:, 3:] + yfrac,
                                    candidates[:, 2:3] + xfrac,
                                    candidates[:, 1:2] - yfrac,
                                ]
                            ),
                            lines_xyxy,
                        ),
                        line_intersect(
                            np.hstack(
                                [
                                    candidates[:, 2:3] + xfrac,
                                    candidates[:, 1:2] - yfrac,
                                    candidates[:, 0:1] - xfrac,
                                    candidates[:, 1:2] - yfrac,
                                ]
                            ),
                            lines_xyxy,
                        ),
                    ),
                ),
            ),
            axis=1,
        )
    )

    non_inside = np.invert(
        np.any(
            np.bitwise_and(
                candidates[:, 0][:, None] - xfrac < lines_xyxy[:, 0],
                np.bitwise_and(
                    candidates[:, 1][:, None] - yfrac < lines_xyxy[:, 1],
                    np.bitwise_and(
                        candidates[:, 2][:, None] + xfrac > lines_xyxy[:, 0],
                        np.bitwise_and(
                            candidates[:, 3][:, None] + yfrac > lines_xyxy[:, 1],
                            np.bitwise_and(
                                candidates[:, 0][:, None] - xfrac < lines_xyxy[:, 2],
                                np.bitwise_and(
                                    candidates[:, 1][:, None] - yfrac
                                    < lines_xyxy[:, 3],
                                    np.bitwise_and(
                                        candidates[:, 2][:, None] + xfrac
                                        > lines_xyxy[:, 2],
                                        candidates[:, 3][:, None] + yfrac
                                        > lines_xyxy[:, 3],
                                    ),
                                ),
                            ),
                        ),
                    ),
                ),
            ),
            axis=1,
        )
    )
    return np.bitwise_and(non_intersecting, non_inside)


def line_intersect(cand_xyxy: np.ndarray, lines_xyxy: np.ndarray) -> np.ndarray:
    """Checks if line segments intersect for all line segments and candidates.

    Args:
        cand_xyxy (np.ndarray): line segments in candidates
        lines_xyxy (np.ndarray): line segments plotted

    Returns:
        np.ndarray: Boolean array with True for non-overlapping candidate segments with line segments.
    """
    intersects = np.bitwise_and(
        ccw(cand_xyxy[:, :2], lines_xyxy[:, :2], lines_xyxy[:, 2:], False)
        != ccw(cand_xyxy[:, 2:], lines_xyxy[:, :2], lines_xyxy[:, 2:], False),
        ccw(cand_xyxy[:, :2], cand_xyxy[:, 2:], lines_xyxy[:, :2], True)
        != ccw(cand_xyxy[:, :2], cand_xyxy[:, 2:], lines_xyxy[:, 2:], True),
    )
    return intersects


def ccw(x1y1: np.ndarray, x2y2: np.ndarray, x3y3: np.ndarray, cand: bool) -> np.ndarray:
    """CCW used in line intersect

    Args:
        x1y1 (np.ndarray):
        x2y2 (np.ndarray):
        x3y3 (np.ndarray):
        cand (bool): using candidate positions (different broadcasting)

    Returns:
        np.ndarray:
    """
    if cand:
        return (
            (-(x1y1[:, 1][:, None] - x3y3[:, 1]))
            * np.repeat(x2y2[:, 0:1] - x1y1[:, 0:1], x3y3.shape[0], axis=1)
        ) > (
            np.repeat(x2y2[:, 1:2] - x1y1[:, 1:2], x3y3.shape[0], axis=1)
            * (-(x1y1[:, 0][:, None] - x3y3[:, 0]))
        )
    return (
        (-(x1y1[:, 1][:, None] - x3y3[:, 1])) * (-(x1y1[:, 0][:, None] - x2y2[:, 0]))
    ) > ((-(x1y1[:, 1][:, None] - x2y2[:, 1])) * (-(x1y1[:, 0][:, None] - x3y3[:, 0])))


def non_overlapping_with_boxes(
    box_arr: np.ndarray, candidates: np.ndarray, xfrac: float, yfrac: float
) -> np.ndarray:
    """Finds candidates not overlapping with allocated boxes.

    Args:
        box_arr (np.ndarray): array with allocated boxes
        candidates (np.ndarray): candidate boxes
        xfrac (float): fraction of the x-dimension to use as margins for text boxes
        yfrac (float): fraction of the y-dimension to use as margins for text boxes

    Returns:
        np.ndarray: Boolean array of shape (K,) with True for non-overlapping candidates with boxes.
    """
    return np.invert(
        np.any(
            np.invert(
                np.bitwise_or(
                    candidates[:, 0][:, None] - xfrac > box_arr[:, 2],
                    np.bitwise_or(
                        candidates[:, 2][:, None] + xfrac < box_arr[:, 0],
                        np.bitwise_or(
                            candidates[:, 1][:, None] - yfrac > box_arr[:, 3],
                            candidates[:, 3][:, None] + yfrac < box_arr[:, 1],
                        ),
                    ),
                )
            ),
            axis=1,
        )
    )


def inside_plot(
    xmin_bound: float,
    ymin_bound: float,
    xmax_bound: float,
    ymax_bound: float,
    candidates: np.ndarray,
) -> np.ndarray:
    """Finds candidates that are inside the plot bounds

    Args:
        xmin_bound (float):
        ymin_bound (float):
        xmax_bound (float):
        ymax_bound (float):
        candidates (np.ndarray): candidate boxes

    Returns:
        np.ndarray: Boolean array of shape (K,) with True for non-overlapping candidates with boxes.
    """
    return np.invert(
        np.bitwise_or(
            candidates[:, 0] < xmin_bound,
            np.bitwise_or(
                candidates[:, 1] < ymin_bound,
                np.bitwise_or(
                    candidates[:, 2] > xmax_bound, candidates[:, 3] > ymax_bound
                ),
            ),
        )
    )



from tqdm import tqdm
from textalloc.non_overlapping_boxes import get_non_overlapping_boxes
import numpy as np
import time
from typing import List, Tuple, Union


def allocate_text(
    fig,
    ax,
    x: Union[np.ndarray, List[float]],
    y: Union[np.ndarray, List[float]],
    text_list: List[str],
    xlims: Tuple[float, float],
    ylims: Tuple[float, float],
    x_scatter: Union[np.ndarray, List[float]] = None,
    y_scatter: Union[np.ndarray, List[float]] = None,
    x_lines: List[Union[np.ndarray, List[float]]] = None,
    y_lines: List[Union[np.ndarray, List[float]]] = None,
    textsize: int = 10,
    distance_margin_fraction: float = 0.015,
    verbose: bool = False,
    draw_lines: bool = False,
    linecolor: str = "k",
    draw_all: bool = False,
    nbr_candidates: int = 0,
):
    """Main function of allocating text-boxes in matplotlib plot

    Args:
        fig (_type_): matplotlib figure used for rendering textbox-sizes.
        ax (_type_): matplotlib axes used for plotting.
        x (Union[np.ndarray, List[float]]): x-coordinates of texts 1d array/list.
        y (Union[np.ndarray, List[float]]): y-coordinates of texts 1d array/list.
        text_list (List[str]): list of texts.
        xlims (Tuple[float, float]): x-limits of plot gotten from ax.get_xlim() (avoids issues with using the function from various interfaces).
        ylims (Tuple[float, float]): y-limits of plot gotten from ax.get_ylim() (avoids issues with using the function from various interfaces).
        x_scatter (Union[np.ndarray, List[float]], optional): x-coordinates of all scattered points in plot 1d array/list. Defaults to None.
        y_scatter (Union[np.ndarray, List[float]], optional): y-coordinates of all scattered points in plot 1d array/list. Defaults to None.
        x_lines (List[Union[np.ndarray, List[float]]], optional): x-coordinates of all lines in plot list of 1d arrays/lists. Defaults to None.
        y_lines (List[Union[np.ndarray, List[float]]], optional): y-coordinates of all lines in plot list of 1d arrays/lists. Defaults to None.
        textsize (int, optional): size of text. Defaults to 10.
        distance_margin_fraction (float, optional): parameter for margins between objects. Increase for larger margins to points and lines. Defaults to 0.015.
        verbose (bool, optional): prints progress using tqdm. Defaults to False.
        draw_lines (bool, optional): draws lines from original points to textboxes. Defaults to False.
        linecolor (str, optional): color code of the lines between points and text-boxes. Defaults to "k".
        draw_all (bool, optional): Draws all texts after allocating as many as possible despit overlap. Defaults to False.
        nbr_candidates (int, optional): Sets the number of candidates used. Defaults to 0.
    """
    t0 = time.time()

    # Ensure good inputs
    assert len(x) == len(y)
    x = np.array(x)
    y = np.array(y)
    if x_scatter is not None:
        assert y_scatter is not None
    if y_scatter is not None:
        assert x_scatter is not None
        assert len(y_scatter) == len(x_scatter)
        x_scatter = np.array(x_scatter)
        y_scatter = np.array(y_scatter)
    if x_lines is not None:
        assert y_lines is not None
    if y_lines is not None:
        assert x_lines is not None
        assert all(
            [len(x_line) == len(y_line) for x_line, y_line in zip(x_lines, y_lines)]
        )
        x_lines = [np.array(x_line) for x_line in x_lines]
        y_lines = [np.array(y_line) for y_line in y_lines]

    # Create boxes in original plot
    if verbose:
        print("Creating boxes")
    original_boxes = []
    for x_coord, y_coord, s in tqdm(zip(x, y, text_list), disable=not verbose):
        ann = ax.text(x_coord, y_coord, s, size=textsize)
        box = ax.transData.inverted().transform(
            ann.get_tightbbox(fig.canvas.get_renderer())
        )
        w, h = box[1][0] - box[0][0], box[1][1] - box[0][1]
        original_boxes.append((x_coord, y_coord, w, h, s))
        ann.remove()

    # Process extracted textboxes
    if verbose:
        print("Processing")
    if x_scatter is None:
        scatterxy = None
    else:
        scatterxy = np.transpose(np.vstack([x_scatter, y_scatter]))
    if x_lines is None:
        lines_xyxy = None
    else:
        lines_xyxy = lines_to_segments(x_lines, y_lines)
    non_overlapping_boxes, overlapping_boxes_inds = get_non_overlapping_boxes(
        original_boxes,
        xlims,
        ylims,
        distance_margin_fraction,
        verbose,
        nbr_candidates,
        scatter_xy=scatterxy,
        lines_xyxy=lines_xyxy,
    )

    # Plot once again
    if verbose:
        print("Plotting")
    for x_coord, y_coord, w, h, s, ind in non_overlapping_boxes:
        ax.text(x_coord, y_coord, s, size=textsize)
        if draw_lines:
            x_near, y_near = find_nearest_point_on_box(
                x_coord, y_coord, w, h, x[ind], y[ind]
            )
            if x_near is not None:
                ax.plot([x[ind], x_near], [y[ind], y_near], linewidth=1, c=linecolor)

    if draw_all:
        for ind in overlapping_boxes_inds:
            ax.text(x[ind], y[ind], text_list[ind], size=textsize)

    if verbose:
        print(f"Finished in {time.time()-t0}s")


def find_nearest_point_on_box(
    xmin: float, ymin: float, w: float, h: float, x: float, y: float
) -> Tuple[float, float]:
    """Finds nearest point on box from point.
    Returns None,None if point inside box

    Args:
        xmin (float): xmin of box
        ymin (float): ymin of box
        w (float): width of box
        h (float): height of box
        x (float): x-coordinate of point
        y (float): y-coordinate of point

    Returns:
        Tuple[float, float]: x,y coordinate of nearest point
    """
    xmax = xmin + w
    ymax = ymin + h
    if x < xmin:
        if y < ymin:
            return xmin, ymin
        elif y > ymax:
            return xmin, ymax
        else:
            return xmin, y
    elif x > xmax:
        if y < ymin:
            return xmax, ymin
        elif y > ymax:
            return xmax, ymax
        else:
            return xmax, y
    else:
        if y < ymin:
            return x, ymin
        elif y > ymax:
            return x, ymax
    return None, None


def lines_to_segments(
    x_lines: List[np.ndarray],
    y_lines: List[np.ndarray],
) -> np.ndarray:
    """Sets up

    Args:
        x_lines (List[np.ndarray]): x-coordinates of all lines in plot list of 1d arrays
        y_lines (List[np.ndarray]): y-coordinates of all lines in plot list of 1d arrays

    Returns:
        np.ndarray: 2d array of line segments
    """
    assert len(x_lines) == len(y_lines)
    n_x_segments = np.sum([len(line_x) - 1 for line_x in x_lines])
    n_y_segments = np.sum([len(line_y) - 1 for line_y in y_lines])
    assert n_x_segments == n_y_segments
    lines_xyxy = np.zeros((n_x_segments, 4))
    iter = 0
    for line_x, line_y in zip(x_lines, y_lines):
        for i in range(len(line_x) - 1):
            lines_xyxy[iter, :] = [line_x[i], line_y[i], line_x[i + 1], line_y[i + 1]]
            iter += 1
    return lines_xyxy



import numpy as np


def generate_candidates(
    w: float,
    h: float,
    x: float,
    y: float,
    xfrac: float,
    yfrac: float,
    nbr_candidates: int,
) -> np.ndarray:
    """Generates 36 candidate boxes

    Args:
        w (float): width of box
        h (float): height of box
        x (float): xmin of box
        y (float): ymin of box
        xfrac (float): fraction of the x-dimension to use as margins for text bboxes
        yfrac (float): fraction of the y-dimension to use as margins for text bboxes
        nbr_candidates (int): nbr of candidates to use. If <1 or >36 uses all 36

    Returns:
        np.ndarray: candidate boxes array
    """
    candidates = np.array(
        [
            [x + xfrac, y + yfrac, x + w + xfrac, y + h + yfrac],  # upper right side
            [x - w - xfrac, y + yfrac, x - xfrac, y + h + yfrac],  # upper left side
            [x - w - xfrac, y - h - yfrac, x - xfrac, y - yfrac],  # lower left side
            [x + xfrac, y - h - yfrac, x + w + xfrac, y - yfrac],  # lower right side
            [x - w - xfrac, y - h / 2, x - xfrac, y + h / 2],  # left side
            [x + xfrac, y - h / 2, x + w + xfrac, y + h / 2],  # right side
            [x - w / 2, y + yfrac, x + w / 2, y + h + yfrac],  # above
            [x - w / 2, y - h - yfrac, x + w / 2, y - yfrac],  # below
            [x - 3 * w / 4, y + yfrac, x + w / 4, y + h + yfrac],  # above left
            [x - w / 4, y + yfrac, x + 3 * w / 4, y + h + yfrac],  # above right
            [x - 3 * w / 4, y - h - yfrac, x + w / 4, y - yfrac],  # below left
            [x - w / 4, y - h - yfrac, x + 3 * w / 4, y - yfrac],  # below right
            # We move all points a bit further from the target
            [
                x + 2 * xfrac,
                y + 2 * yfrac,
                x + w + 2 * xfrac,
                y + h + 2 * yfrac,
            ],  # upper right side
            [
                x - w - 2 * xfrac,
                y + 2 * yfrac,
                x - 2 * xfrac,
                y + h + 2 * yfrac,
            ],  # upper left side
            [
                x - w - 2 * xfrac,
                y - h - 2 * yfrac,
                x - 2 * xfrac,
                y - 2 * yfrac,
            ],  # lower left side
            [
                x + 2 * xfrac,
                y - h - 2 * yfrac,
                x + w + 2 * xfrac,
                y - 2 * yfrac,
            ],  # lower right side
            [x - w - 2 * xfrac, y - h / 2, x - 2 * xfrac, y + h / 2],  # left side
            [x + 2 * xfrac, y - h / 2, x + w + 2 * xfrac, y + h / 2],  # right side
            [x - w / 2, y + 2 * yfrac, x + w / 2, y + h + 2 * yfrac],  # above
            [x - w / 2, y - h - 2 * yfrac, x + w / 2, y - 2 * yfrac],  # below
            [x - 3 * w / 4, y + 2 * yfrac, x + w / 4, y + h + 2 * yfrac],  # above left
            [x - w / 4, y + 2 * yfrac, x + 3 * w / 4, y + h + 2 * yfrac],  # above right
            [x - 3 * w / 4, y - h - 2 * yfrac, x + w / 4, y - 2 * yfrac],  # below left
            [x - w / 4, y - h - 2 * yfrac, x + 3 * w / 4, y - 2 * yfrac],  # below right
            [
                x + 3 * xfrac,
                y + 3 * yfrac,
                x + w + 3 * xfrac,
                y + h + 3 * yfrac,
            ],  # upper right side
            [
                x - w - 3 * xfrac,
                y + 3 * yfrac,
                x - 3 * xfrac,
                y + h + 3 * yfrac,
            ],  # upper left side
            [
                x - w - 3 * xfrac,
                y - h - 3 * yfrac,
                x - 3 * xfrac,
                y - 3 * yfrac,
            ],  # lower left side
            [
                x + 3 * xfrac,
                y - h - 3 * yfrac,
                x + w + 3 * xfrac,
                y - 3 * yfrac,
            ],  # lower right side
            [x - w - 3 * xfrac, y - h / 2, x - 3 * xfrac, y + h / 2],  # left side
            [x + 3 * xfrac, y - h / 2, x + w + 3 * xfrac, y + h / 2],  # right side
            [x - w / 2, y + 3 * yfrac, x + w / 2, y + h + 3 * yfrac],  # above
            [x - w / 2, y - h - 3 * yfrac, x + w / 2, y - 3 * yfrac],  # below
            [x - 3 * w / 4, y + 3 * yfrac, x + w / 4, y + h + 3 * yfrac],  # above left
            [x - w / 4, y + 3 * yfrac, x + 3 * w / 4, y + h + 3 * yfrac],  # above right
            [x - 3 * w / 4, y - h - 3 * yfrac, x + w / 4, y - 3 * yfrac],  # below left
            [x - w / 4, y - h - 3 * yfrac, x + 3 * w / 4, y - 3 * yfrac],  # below right
        ]
    )
    if nbr_candidates > 0 and nbr_candidates < len(candidates):
        return candidates[:nbr_candidates, :]
    return candidates
